# -*- coding: utf-8 -*-
# author:shihua
# coder:shihua
# 这是一个meoteofence对外接口封装类，主要功能对外提供封装好接口。
"""
模块介绍
-------

这是一个meoteofence对外接口封装类，主要功能对外提供封装好接口。

设计模式：

    （1）  无 

关键点：    

    （1）接口封装

主要功能：            

    （1）对外提供封装接口                          

使用示例
-------


类说明
------

"""



####### 载入程序包 ##########################################################
############################################################################



import pluggy
from meoteofence.hook_specs import PretreatmentSpec
from meoteofence.hook_impl import PretreatmentPlugin
from meoteofence import HOOK_NAMESPACE
from meoteofence import hdf5_read_ndarray
import time
import os
from multiprocessing import Pool
import pandas as pd
import numpy as np
from scipy.interpolate import interp1d



####### 对外接口封装类 ######################################################
### 设计模式：                                                           ###
### （1）无                                                              ###
### 关键点：                                                             ###
### （1）接口封装                                                        ###
### 主要功能：                                                           ###
### （1）对外提供封装接口                                                  ###
############################################################################



####### 整理的二级API ###########################################################################################
###############################################################################################################



### 整理一个二级API
def secondary_api_meoteo_pretreatment(file_path,variable_name,station_location,hdf5_store_path_pre,resolution_ratio,grid_size):
    '''
    方法功能：

        定义一个气象预处理的二级API

    参数：
        file_path (str): 预处理数据目标文件夹
        hdf5_store_path_pre (str): hdf5存储路径前缀
        station_location (tuple): 场站坐标元组
        variable_name_list (list): 气象物理量列表
        resolution_ratio (float): 分辨率
        grid_size (int): 网格大小

    返回：
        result (str): 运行结果信息
    '''

    ### 创建hook插件管理对象
    pm = pluggy.PluginManager(HOOK_NAMESPACE)
    ### 添加气象预处理接口
    pm.add_hookspecs(PretreatmentSpec)
    ### 注册气象预处理插件
    pm.register(PretreatmentPlugin())
    result = pm.hook.meoteo_pretreatment(file_path=file_path,
                                         variable_name=variable_name,
                                         station_location=station_location,
                                         hdf5_store_path_pre=hdf5_store_path_pre,
                                         resolution_ratio=resolution_ratio,
                                         grid_size=grid_size)
    print(result)

    return result



####### 对外接口封装类 ########################################################################################
#############################################################################################################



class Meoteodata(object):
    '''
    类介绍:

        这是一个气象数据对象类,主要功能提供气象相关处理操作
    '''


    def __init__(self,base_folder_path,hdf5_store_path_pre,station_location,variable_name_list,resolution_ratio,grid_size,n_jobs):
        '''
        属性功能:

            定义一个初始化属性的功能方法

        参数:
            base_folder_path (str): 预处理数据目标文件夹
            hdf5_store_path_pre (str): hdf5存储路径前缀
            station_location (tuple): 场站坐标元组
            variable_name_list (list): 气象物理量列表
            resolution_ratio (float): 分辨率
            grid_size (int): 网格大小
            n_jobs (int): 并行个数
        '''

        self.base_folder_path = base_folder_path
        self.file_path_list = os.listdir(self.base_folder_path)
        self.hdf5_store_path_pre = hdf5_store_path_pre
        self.station_location = station_location
        self.variable_name_list = variable_name_list
        self.resolution_ratio = resolution_ratio
        self.grid_size = grid_size
        self.n_jobs = n_jobs


    def run(self):
        '''
        属性功能：

            定义一个初始化属性的方法

        参数：
            base_folder_path (str): 预处理数据目标文件夹
            hdf5_store_path_pre (str): hdf5存储路径前缀
            station_location (tuple): 场站坐标元组
            variable_name_list (list): 气象物理量列表
            resolution_ratio (float): 分辨率
            grid_size (int): 网格大小
            n_jobs (int): 并行个数

        返回：
            result (str): 运行结果信息,包含运行时间信息
        '''


        ### 配置初始参数
        base_folder_path = self.base_folder_path
        file_path_list = self.file_path_list
        hdf5_store_path_pre = self.hdf5_store_path_pre
        station_location = self.station_location
        variable_name_list = self.variable_name_list
        resolution_ratio = self.resolution_ratio
        grid_size = self.grid_size
        n_jobs = self.n_jobs
        ### 使用multiprocessing进行并行执行
        ### 开始执行(时间并行方式运行)
        start_time = time.time()
        ### 使用进程池循环
        for variable_name in variable_name_list:
            ### 创建pool对象
            pool = Pool(n_jobs) 
            print(variable_name)
            for file in file_path_list:
                file_path = base_folder_path + file
                ### 向进程池队列添加事件
                pool.apply_async(func=secondary_api_meoteo_pretreatment,args=(file_path,variable_name,station_location,hdf5_store_path_pre,resolution_ratio,grid_size))
            ### 关闭进程池
            pool.close()
            ### 回收进程池
            pool.join()
        end_time = time.time()
        parallel_runtime = end_time - start_time
        result = 'meoteodata run well done! Total time is {}'.format(parallel_runtime)

        return result



####### 工具函数集合 ###########################################################################################################################################################################
##############################################################################################################################################################################################



### 查询工具函数
def data_drill_up(hdf5_store_path_pre,variable_name,level_name,forecast_time):
    '''
    方法功能：

        定义一个从指定日期HDF5文件夹中进行数据上钻查询的方法，以预报时间为行索引，以变量和层级组合为列索引。

    参数：
        hdf5_store_path_pre (str): hdf5存储路径前缀
        variable_name_list (str): 气象物理量
        level_name (str): 层级名称
        forecast_time (int): 预报时间长度

    返回：
        tmp_concat_df (DataFrame): 组合拼接好的数据集合
    '''

    ### 开始第一层循环----预报时间
    ### 创建一个初始空dataframe用于concat
    tmp_concat_df = pd.DataFrame()
    for i in range(0,forecast_time,1):
        if len(str(i)) == 1:
            h5file_name = 'f0' + str(i)
        else:
            h5file_name = 'f' + str(i)
        # print(h5file_name)
        ### 从指定h5中读取对应ndarray数据
        tmp_ndarray = hdf5_read_ndarray(h5path='{}{}.h5'.format(hdf5_store_path_pre,h5file_name),var_name=variable_name,level_name=level_name) # level_100_m
        # print(tmp_ndarray)
        ### 获取网格个数，以供后续使用
        length = list(tmp_ndarray.shape)[0] * list(tmp_ndarray.shape)[1]
        # print(length)
        ### 将目标矩阵拉平，并转换为列表
        tmp_reshape_ndarray = tmp_ndarray.reshape(1,length)
        # print(tmp_reshape_ndarray)
        ### 按照宽表构建dataframe
        tmp_df = pd.DataFrame(tmp_reshape_ndarray)
        tmp_concat_df = pd.concat([tmp_concat_df,tmp_df],axis=0) ### 选择axis参数0,按行合并
    ### 重置索引
    tmp_concat_df = tmp_concat_df.reset_index(drop=True)
    ### 重新整理列名----var-level-location
    ### 创建一个空列表用于重新命名列名
    tmp_column_name_list = []
    ### 重新整理列名并添加到tmp_column_name_list中
    for i in range(0,length,1):
        tmp_column_name = str(variable_name) + ':' + str(level_name) + ':' + str(i)
        tmp_column_name_list.append(tmp_column_name)
    ### 重置tmp_concat_df列名
    tmp_concat_df.columns = tmp_column_name_list
    # ### 创建时间列
    # tmp_time_list = [i+1 for i in range(0,int(forecast_time),1)]
    # ### 向tmp_concat_df添加时间列
    # tmp_concat_df['time'] = tmp_time_list
    # print('-------------------------------------------------------------------------------------------------------------------------------------------------')
    # print(tmp_concat_df)
    # print(tmp_concat_df.time)
    result = 'Data concat well done!'
    print(result)

    return tmp_concat_df



### 数据插值工具函数
def data_interpolate(tmp_concat_df,index_delta,kind):
    '''
    函数功能：

        定义一个针对气象数据进行插值的函数

    参数：
        tmp_concat_df (dataframe): 拼接好的数据集合
        index_delta (float): 索引间隔
        kind (str): 插值方法

    返回：
        tmp_interpolate_df (dataframe): 插值完成的数据集合
    '''

    ### 创建一个空dataframe用来装插值后的变量列表
    tmp_interpolate_df = pd.DataFrame()
    ### 开始循环----dataframe列变量
    first_column_name = tmp_concat_df.columns[0]
    ### 选取变量
    tmp_array = tmp_concat_df[first_column_name].values
    # print(tmp_array,type(tmp_array))
    ### 确定原本的x值
    tmp_original_x = np.array(list(tmp_concat_df.index))
    for tmp_column_name in tmp_concat_df.columns:
        # print(tmp_column_name)
        ### 对数据进行插值
        ### 拟合曲线
        tmp_func = interp1d(tmp_original_x,tmp_array,kind=kind)
        ### 用拟合好的曲线在新的序列上计算新的值
        start_index = tmp_concat_df.index[0]
        end_index = tmp_concat_df.index[-1]
        # index_delta = 0.25
        tmp_new_x = np.arange(start_index,end_index+index_delta,index_delta)
        tmp_new_array = tmp_func(tmp_new_x)
        tmp_interpolate_df[tmp_column_name] = list(tmp_new_array)

    return tmp_interpolate_df



############################################################################################################################################################################################
############################################################################################################################################################################################





from meoteofence import hdf5_read_ndarray
import pandas as pd
import numpy as np



### 参数配置
### 自己手动制定需要的变量和层级和目标文件夹
hdf5_store_path_pre = './data_handled/20221122/'
variable_name = 'Wind speed (gust)' # 'Temperature'
level_name = 'level_0' # level_100_M
### 设置预报时间
forecast_time = 24



### 开始第一层循环----预报时间
### 创建一个初始空dataframe用于concat
tmp_concat_df = pd.DataFrame()
for i in range(0,forecast_time,1):
    if len(str(i)) == 1:
        h5file_name = 'f0' + str(i)
    else:
        h5file_name = 'f' + str(i)
    print(h5file_name)
    ### 从指定h5中读取对应ndarray数据
    tmp_ndarray = hdf5_read_ndarray(h5path='{}{}.h5'.format(hdf5_store_path_pre,h5file_name),var_name=variable_name,level_name=level_name) # level_100_m
    print(tmp_ndarray)
    ### 获取网格个数，以供后续使用
    length = list(tmp_ndarray.shape)[0] * list(tmp_ndarray.shape)[1]
    print(length)
    ### 将目标矩阵拉平，并转换为列表
    tmp_reshape_ndarray = tmp_ndarray.reshape(1,length)
    print(tmp_reshape_ndarray)
    ### 按照宽表构建dataframe
    tmp_df = pd.DataFrame(tmp_reshape_ndarray)
    tmp_concat_df = pd.concat([tmp_concat_df,tmp_df],axis=0) ### 选择axis参数0,按行合并
### 重置索引
tmp_concat_df = tmp_concat_df.reset_index(drop=True)
### 重新整理列名----var-level-location
### 创建一个空列表用于重新命名列名
tmp_column_name_list = []
### 重新整理列名并添加到tmp_column_name_list中
for i in range(0,length,1):
    tmp_column_name = str(variable_name) + ':' + str(level_name) + ':' + str(i)
    tmp_column_name_list.append(tmp_column_name)
### 重置tmp_concat_df列名
tmp_concat_df.columns = tmp_column_name_list
# ### 创建时间列
# tmp_time_list = [i+1 for i in range(0,int(forecast_time),1)]
# ### 向tmp_concat_df添加时间列
# tmp_concat_df['time'] = tmp_time_list
print('-------------------------------------------------------------------------------------------------------------------------------------------------')
print(tmp_concat_df)
# print(tmp_concat_df.time)







from meoteofence import get_initial_info_from_file,convert_matrix_to_list,get_coordinate_pairs_according_to_station,form_base_coordinate,combine_variable_and_levels,hdf5_write_ndaray,hdf5_read_ndarray,obtain_data_and_cache_to_hdf5
import pygrib 
import numpy as np
import os
import time

### 使用pygrib加载目标文件中的分层数据对象
file_path='./data/20221122/gfs.pgrb2.f19.grib2'
grbs = pygrib.open(file_path)
grbs.seek(0)
for grb in grbs:
    print(grb)


# ### 获取经纬度列表和坐标点位
# file_path='./data/20221122/gfs.pgrb2.f19.grib2'
# variable_name = 'Temperature' # Wind speed (gust) # Relative humidity # Total Cloud Cover
# locations,variable_level_list,grbs_obj = get_initial_info_from_file(file_path=file_path,variable_name=variable_name)


# ### 将经纬度坐标转换为两个列表
# latitude_list,longitude_list,longtitude_latitude_list = convert_matrix_to_list(locations=locations)
# # print(latitude_list,longitude_list,longtitude_latitude_list)


# ### 以场站坐标为依据从两个列表中选取坐标对，示例使用longtitude-31.23,latitude-90.24
# station_location = (31.23,90.24)
# upleft_standard_location,resolution_ratio = get_coordinate_pairs_according_to_station(longitude_list=longitude_list,latitude_list=latitude_list,station_location=station_location)
# print(upleft_standard_location,resolution_ratio)


# ### 对应形成坐标矩阵
# upleft_grid_location,upright_grid_location,downleft_grid_location,downright_grid_location= form_base_coordinate(upleft_standard_location=upleft_standard_location,resolution_ratio=0.25,grid_size=6)
# print(upleft_grid_location,upright_grid_location,downleft_grid_location,downright_grid_location)


# ### 组合变量与层级得到宽表字段
# variable_level_combine_list = combine_variable_and_levels(variable_name=variable_name,variable_level_list=variable_level_list)
# print(variable_level_combine_list)


# ### 获取数据并缓存到HDF5中
# obtain_data_and_cache_to_hdf5(grbs_obj=grbs_obj,
#                               variable_name=variable_name,
#                               variable_level_list=variable_level_list,
#                               hdf5_name='./data_handled/20221122/gfs.pgrb2.f19.h5',
#                               upleft_grid_location=upleft_grid_location,
#                               upright_grid_location=upright_grid_location,
#                               downleft_grid_location=downleft_grid_location,
#                               downright_grid_location=downright_grid_location)


# ### 从指定h5中读取对应ndarray数据
# tmp_ndarray = hdf5_read_ndarray(h5path='./data_handled/20221122/gfs.pgrb2.f00.h5',var_name=variable_name,level_name='level_100_m')
# print(tmp_ndarray)

### 整合成函数
def meoteo_pretreatment(file_path,variable_name,station_location):
    ### 获取经纬度列表和坐标点位
    locations,variable_level_list,grbs_obj = get_initial_info_from_file(file_path=file_path,variable_name=variable_name)
    ### 将经纬度坐标转换为两个列表
    latitude_list,longitude_list,longtitude_latitude_list = convert_matrix_to_list(locations=locations)
    # print(latitude_list,longitude_list,longtitude_latitude_list)
    ### 以场站坐标为依据从两个列表中选取坐标对，示例使用longtitude-31.23,latitude-90.24
    # station_location = (31.23,90.24)
    upleft_standard_location,resolution_ratio = get_coordinate_pairs_according_to_station(longitude_list=longitude_list,latitude_list=latitude_list,station_location=station_location)
    # print(upleft_standard_location,resolution_ratio)
    ### 对应形成坐标矩阵
    upleft_grid_location,upright_grid_location,downleft_grid_location,downright_grid_location= form_base_coordinate(upleft_standard_location=upleft_standard_location,resolution_ratio=0.25,grid_size=6)
    # print(upleft_grid_location,upright_grid_location,downleft_grid_location,downright_grid_location)
    ### 组合变量与层级得到宽表字段
    variable_level_combine_list = combine_variable_and_levels(variable_name=variable_name,variable_level_list=variable_level_list)
    # print(variable_level_combine_list)
    ### 获取数据并缓存到HDF5中
    ### 配置HDF5文件路径
    hdf5_file_path = file_path.split('.')[-2]+'.'+'h5'
    hdf5_file_path = hdf5_store_path_pre + hdf5_file_path
    print(hdf5_file_path)
    ### 开始向HDF5缓存
    obtain_data_and_cache_to_hdf5(grbs_obj=grbs_obj,
                                variable_name=variable_name,
                                variable_level_list=variable_level_list,
                                hdf5_name=hdf5_file_path,
                                upleft_grid_location=upleft_grid_location,
                                upright_grid_location=upright_grid_location,
                                downleft_grid_location=downleft_grid_location,
                                downright_grid_location=downright_grid_location)    
    print('======================================================>>>>>>',file)






### 根据预报时间开始顺序循环运行并统计时间
start_time = time.time()
### 配置初始参数
file_path_list = os.listdir('./data/20221122')
hdf5_store_path_pre = './data_handled/20221122/'
station_location = (31.23,90.24)
### 遍历变量列表处理
# variable_name = 'Wind speed (gust)' #'Temperature'
for variable_name in ['Temperature','Relative humidity','Total Cloud Cover','Wind speed (gust)']:
    print(variable_name)
    for file in file_path_list:
        file_path = './data/20221122/' + file
        meoteo_pretreatment(file_path=file_path,variable_name=variable_name,station_location=station_location)
end_time = time.time()
print('*************** total time is {}'.format(end_time-start_time))






### 从指定h5中读取对应ndarray数据
variable_name = 'Wind speed (gust)'#'Temperature' # 'Wind speed (gust)'
tmp_ndarray = hdf5_read_ndarray(h5path='./data_handled/20221122/f23.h5',var_name=variable_name,level_name='level_0') # level_100_m
print(tmp_ndarray)



from meoteofence.interface import Meoteodata,data_drill_up,data_interpolate
from meoteofence import hdf5_read_ndarray
import pandas as pd
import numpy as np



# ### 第一步：原始气象数据抽取解析（一般一天执行一次，没有重复执行必要，可以跑批的方式预先执行完成）
# ### 一次执行多个文件夹
# folders = ['20221123','20221124']
# for tmp_folder in folders:
#     ### 创建气象数据对象
#     ### 配置初始参数，生成气象数据处理对象
#     meoteodata = Meoteodata(base_folder_path='./data/{}/'.format(tmp_folder),
#                             hdf5_store_path_pre = './data_handled/{}/'.format(tmp_folder),
#                             station_location = (35.23,95.24),
#                             variable_name_list = ['Temperature','Relative humidity','Total Cloud Cover','Wind speed (gust)'],
#                             resolution_ratio = 0.25,
#                             grid_size = 6,
#                             n_jobs = 24)
#     ### 开始执行气象预处理程序
#     result = meoteodata.run()
#     # result = meoteodata.run(func=secondary_api_meoteo_pretreatment) 
#     print(result)


### 第二步：从气象抽取解析结果中进行上钻操作，并整合出96时间点位的气象物理量（该步骤为频繁使用步骤，基于HDF5保证数据读取速度）
### 测试结果
# ### 从指定h5中读取对应ndarray数据（此步骤为一级API，一般不使用，为更高级自定义操作提供支持）
# variable_name = 'Wind speed (gust)'#'#'Temperature' # 'Wind speed (gust)'
# tmp_ndarray = hdf5_read_ndarray(h5path='./data_handled/20221124/f23.h5',var_name=variable_name,level_name='level_0') # level_100_m
# print(tmp_ndarray)


### 根据变量和层级对指定文件夹下的数据进行上钻操作
tmp_concat_df_a = data_drill_up(hdf5_store_path_pre='./data_handled/20221123/',
                                 variable_name = 'Wind speed (gust)',
                                 level_name = 'level_0',
                                 forecast_time = 24)
tmp_concat_df_b = data_drill_up(hdf5_store_path_pre='./data_handled/20221124/',
                                 variable_name = 'Wind speed (gust)',
                                 level_name = 'level_0',
                                 forecast_time = 24)
tmp_concat_df = pd.concat([tmp_concat_df_a,tmp_concat_df_b],axis=0)
tmp_concat_df = tmp_concat_df.reset_index(drop=True)


### 对目标数据集合进行内插值
tmp_interpolate_df = data_interpolate(tmp_concat_df=tmp_concat_df,index_delta=0.25,kind='linear')
tmp_target_df = tmp_interpolate_df.iloc[:96,:]
print(tmp_target_df)






import pluggy
from meoteofence.hook_specs import PretreatmentSpec
from meoteofence.hook_impl import PretreatmentPlugin
from meoteofence import HOOK_NAMESPACE
from meoteofence import hdf5_read_ndarray
import time
import os
from multiprocessing import Pool

# ### 创建hook插件管理对象
# pm = pluggy.PluginManager(HOOK_NAMESPACE)
# ### 添加气象预处理接口
# pm.add_hookspecs(PretreatmentSpec)
# ### 注册气象预处理插件
# pm.register(PretreatmentPlugin())
# ### 开始执行(顺序循环方式运行)
# start_time = time.time()
# ### 配置初始参数
# file_path_list = os.listdir('./data/20221122')
# hdf5_store_path_pre = './data_handled/20221122/'
# station_location = (31.23,90.24)
# variable_name_list = ['Temperature','Relative humidity','Total Cloud Cover','Wind speed (gust)']
# ### 使用for循环
# for variable_name in variable_name_list:
#     print(variable_name)
#     for file in file_path_list:
#         file_path = './data/20221122/' + file
#         result = pm.hook.meoteo_pretreatment(file_path=file_path,
#                                              variable_name=variable_name,
#                                              station_location=station_location,
#                                              hdf5_store_path_pre=hdf5_store_path_pre,
#                                              resolution_ratio=0.25,
#                                              grid_size=6)
# print(result)
# end_time = time.time()
# sequential_runtime = end_time - start_time
# print('-------------------------- total time is {}'.format(sequential_runtime))






# def test_func():
## 使用multiprocessing进行并行执行
## 整理一个二级API
def secondary_api_meoteo_pretreatment(file_path,variable_name,station_location,hdf5_store_path_pre,resolution_ratio,grid_size):
    ### 创建hook插件管理对象
    pm = pluggy.PluginManager(HOOK_NAMESPACE)
    ### 添加气象预处理接口
    pm.add_hookspecs(PretreatmentSpec)
    ### 注册气象预处理插件
    pm.register(PretreatmentPlugin())
    ### 开始执行气象预处理程序
    result = pm.hook.meoteo_pretreatment(file_path=file_path,
                                            variable_name=variable_name,
                                            station_location=station_location,
                                            hdf5_store_path_pre=hdf5_store_path_pre,
                                            resolution_ratio=resolution_ratio,
                                            grid_size=grid_size)
    print(result)

    return result


folder_path = './data/20221122/'
hdf5_store_path_pre = './data_handled/20221122/'
station_location = (31.23,90.24)
variable_name_list = ['Temperature','Relative humidity','Total Cloud Cover','Wind speed (gust)']
resolution_ratio = 0.25
grid_size = 6
n_jobs = 4

def test_func(folder_path,hdf5_store_path_pre,station_location,variable_name_list,resolution_ratio,grid_size,n_jobs):
    ### 开始执行(时间并行方式运行)
    start_time = time.time()
    ### 配置初始参数
    # folder_path = './data/20221122/'
    file_path_list = os.listdir(folder_path)
    # hdf5_store_path_pre = './data_handled/20221122/'
    # station_location = (31.23,90.24)
    # variable_name_list = ['Temperature','Relative humidity','Total Cloud Cover','Wind speed (gust)']
    # resolution_ratio = 0.25
    # grid_size = 6
    # n_jobs = 4
    ### 使用进程池循环
    for variable_name in variable_name_list:
        ### 创建pool对象
        pool = Pool(n_jobs) 
        print(variable_name)
        for file in file_path_list:
            file_path = folder_path + file
            ### 向进程池队列添加事件
            pool.apply_async(func=secondary_api_meoteo_pretreatment,args=(file_path,variable_name,station_location,hdf5_store_path_pre,resolution_ratio,grid_size))
        ### 关闭进程池
        pool.close()
        ### 回收进程池
        pool.join()
    end_time = time.time()
    parallel_runtime = end_time - start_time
    print('-------------------------- total time is {}'.format(parallel_runtime))### 开始执行(时间并行方式运行)





test_func(folder_path,hdf5_store_path_pre,station_location,variable_name_list,resolution_ratio,grid_size,n_jobs)
### 从指定h5中读取对应ndarray数据
variable_name = 'Wind speed (gust)'#'Temperature' # 'Wind speed (gust)'
tmp_ndarray = hdf5_read_ndarray(h5path='./data_handled/20221122/f23.h5',var_name=variable_name,level_name='level_0') # level_100_m
print(tmp_ndarray)





# -*- coding: utf-8 -*-
# author:shihua
# coder:shihua
# 这是一个包入口文件，主要功能收集各种工具函数，主要技术__all__暴露指定函数
"""
模块介绍
-------

这是一个包入口文件，主要功能收集各种工具函数，主要技术__all__暴露指定函数

设计模式：

    （1）  无 

关键点：    

    （1）__all__

主要功能：            

    （1）基础操作工具集合                                

使用示例
-------


类说明
------

"""



####### 载入程序包 ##########################################################
############################################################################



import pygrib
from tables import *
import numpy as np
import pandas as pd



####### 基础工具集合类 ######################################################
### 设计模式：                                                           ###
### （1）无                                                              ###
### 关键点：                                                             ###
### （1）基础依赖                                                        ###
### 主要功能：                                                           ###
### （1）基础操作工具集合                                                 ###
############################################################################


####### 基础工具集合 #####################################################################################
########################################################################################################



### 暴露指定的公开接口
__all__ = ['get_coordinate_matrix_from_file',
           'convert_matrix_to_list',
           'get_coordinate_pairs_according_to_station',
           'form_base_coordinate',
           'combine_variable_and_levels',
           'hdf5_write_ndaray',
           'hdf5_read_ndarray',
           'obtain_data_and_cache_to_hdf5',
           'HOOK_NAMESPACE']



HOOK_NAMESPACE = 'meoteofence'



def get_initial_info_from_file(file_path,variable_name):
    '''
    函数功能：

        定义一个获取文件经纬度的函数

    参数：
        file_path (str): 文件路径
        variable_name (str): 变量名称

    返回：
        locations (object): 坐标点位信息
        variable_level_list (list): 层级列表
        grbs_obj (object): grbs对象
    ''' 

    ### 使用pygrib加载目标文件中的分层数据对象
    grbs = pygrib.open(file_path)
    grbs.seek(0)
    ### 缓存grbs对象
    grbs_obj = grbs.select(name=variable_name)
    ### 根据变量来获取经纬度表
    locations = grbs_obj[0].latlons()
    ### 根据变量来获取层级列表
    variable_level_list = []
    ### 获取grbs层级
    levels = len(grbs.select(name=variable_name))
    ### 遍历grbs以获取层级列表
    for i in range(0,levels,1):
        grb = grbs_obj[i]
        grb_str = str(grb)
        start_index = grb_str.find('level')
        sub_grb_str = grb_str[start_index:]
        sub_grb_str_split = sub_grb_str.split(':')
        variable_level_list.append(sub_grb_str_split[0].replace(' ','_'))

    return locations,variable_level_list,grbs_obj



def convert_matrix_to_list(locations):
    '''
    函数功能：

        定义一个将坐标信息转换为两个列表和坐标元组列表的函数

    参数：
        locations (object): pygrib获取的坐标信息

    返回：
        longitude (list): 纬度列表
        latitude (list): 经度列表
        longitude_latitude_list (list): 经纬度列表    
    '''

    ### 对经纬度进行处理，得到坐标点位数据对
    latitude_array = np.array(locations[0])
    longitude_array = np.array(locations[1])
    latitude_list = [list(set(latitude_array[i]))[0] for i in range(0,latitude_array.shape[0])]
    longitude_list = list(longitude_array[0])
    longitude_latitude_list = [(i,j) for i in latitude_list for j in longitude_list]

    return latitude_list,longitude_list,longitude_latitude_list



def get_coordinate_pairs_according_to_station(longitude_list,latitude_list,station_location):
    '''
    函数功能：

        定义一个以场站坐标为依据从两列表中选取坐标对的函数

    参数：
        longitude_list (list): 经度列表
        latitude_list (list): 纬度列表
        station_location (tuple): 场站坐标元组

    返回：
        upleft_standard_location (tuple): 选取的左上标准点位
        resolution_ratio (float): 网格分辨率
    '''

    ### 选取左上标准点位
    chanced_longitude_list = [i for i in longitude_list if i>=station_location[1]]
    chanced_latitude_list = [i for i in latitude_list if i>=station_location[0]]
    upleft_standard_location = (chanced_latitude_list[0],chanced_longitude_list[0])
    ### 计算网格分辨率
    resolution_ratio = chanced_latitude_list[1] - chanced_latitude_list[0]

    return upleft_standard_location,resolution_ratio



def form_base_coordinate(upleft_standard_location,resolution_ratio,grid_size):
    '''
    函数功能：

        定义一个对应形成矩阵的函数

    参数：
        upleft_standard_location (tuple): 左上标准点位
        resolution_ratio (float): 网格分辨率
        grid_size (int): 网格大小

    返回：
        upleft_grid_location (tuple): 目标网格左上点位
        upright_grid_location (tuple): 目标网格右上点位
        downleft_grid_location (tuple): 目标网格左下点位
        downright_grid_location (tuple): 目标网格右下点位
    '''

    ### 根据网格大小确定划分步长
    if grid_size % 2 == 0: ### 取余为0则为偶数
        up_step = grid_size / 2 -1
        left_step = grid_size / 2 -1
    else: ### 其他情况则为奇数
        up_step = int((grid_size-1) / 2)
        left_step = int((grid_size-1) / 2)
    ### 根据步长先锁定目标网格左上点位，然后根据网格大小参数确定目标网格的其余四点位
    upleft_grid_location = (upleft_standard_location[0]+up_step*resolution_ratio,upleft_standard_location[1]+left_step*resolution_ratio)
    upright_grid_location = (upleft_grid_location[0],upleft_grid_location[1]+(grid_size-1)*resolution_ratio)
    downleft_grid_location = (upleft_grid_location[0]-(grid_size-1)*resolution_ratio,upleft_grid_location[1])
    downright_grid_location = (upleft_grid_location[0]-(grid_size-1)*resolution_ratio,upleft_grid_location[1]+(grid_size-1)*resolution_ratio)

    return upleft_grid_location,upright_grid_location,downleft_grid_location,downright_grid_location
    


def combine_variable_and_levels(variable_name,variable_level_list):
    '''
    函数功能：

        定义一个组合变量与层级的函数

    参数：
        variable_name (str): 变量名称
        variable_level_list (list): 层级列表

    返回：
        variable_level_combine_list (list): 变量层级组合列表
    '''

    variable_level_combine_list = [variable_name+'_'+tmp_level for tmp_level in variable_level_list]

    return variable_level_combine_list



def hdf5_write_ndaray(h5path,var_name,level_name,ndarray_obj):
    '''
    函数功能：

        定义一个将ndarray写入H5文件的方法

    参数：
        h5path (str): hdf5文件路径
        var_name (str): 气象物理量名称，对应HDF5的组名称
        level_name (str): 气象空间维度层级名称，对应HDF5的数据集合名称
        ndarray_obj (numpy.ndarray): numpy的ndarray矩阵数据对象

    返回：
        result (str): 运行结果信息
    '''

    ### 打开文件流，以追加的模式打开
    h5file = open_file(filename=h5path,mode='a')
    ### 创建数据组
    try:
        data_group = h5file.create_group(where="/",name=var_name,title="meoteovar is {}".format(var_name))
    except:
        data_group = "/{}".format(var_name)
    ### 压缩算法选择
    filters = Filters(complevel=5,complib='blosc')
    ### 在数据组下写入数据对象
    try:
        h5file.create_earray(where=data_group,name=level_name,obj=ndarray_obj,filters=filters)
    except:
        pass
    ### 关闭文件流
    h5file.close()
    result = 'Hdf5 write data well done!'
    print(result)

    return result



def hdf5_read_ndarray(h5path,var_name,level_name):
    '''
    函数功能：

        定义一个hdf5读取数据的方法

    参数：
        h5path (str): hdf5文件路径
        var_name (str): 气象物理量名称，对应HDF5的组名称
        level_name (str): 气象空间维度层级名称，对应HDF5的数据集合名称

    返回：
        ndarray_obj (numpy.ndarray): numpy的ndarray矩阵数据对象
    '''

    ### 打开文件流
    h5file = open_file(filename=h5path,mode='r')
    ### 获取数据索引
    data_reader = h5file.get_node(where='/{}'.format(var_name),name=level_name)
    ### 获取数据帧
    data_ndarray = data_reader.read()
    ### 关闭文件流
    h5file.close()
    result = 'HDF5 read ndarray well done!'
    print(result)

    return data_ndarray



def obtain_data_and_cache_to_hdf5(grbs_obj,variable_name,variable_level_list,hdf5_name,upleft_grid_location,upright_grid_location,downleft_grid_location,downright_grid_location):
    '''
    函数功能：

        定义一个获取数据并缓存到HDF5中的函数

    参数：
        grbs_obj (object): grbs对象
        variable_name (str): 气象物理量名称
        variable_level_list (list): 变量层级列表
        hdf5_name (str): hdf5文件名称
        upleft_grid_location (tuple): 左上基准点位
        upright_grid_location (tuple): 右上基准点位
        downleft_grid_location (tuple): 左下基准点位
        downright_grid_location (tuple): 右下基准点位

    返回：
        result (str): 运行结果信息
    '''

    ### 根据变量层级遍历获取数据，grbs_obj为给定变量的grbs查询对象
    for i,tmp_var_level in enumerate(variable_level_list):
        print(i,tmp_var_level)
        grb = grbs_obj[i]
        tmp = grb.data(lat1=downleft_grid_location[0],lat2=upleft_grid_location[0],lon1=upleft_grid_location[1],lon2=upright_grid_location[1])
        ### 索引选择0,表示选择对应网格上的物理量矩阵
        # print(tmp[0],type(tmp[0]))
        # print(tmp[0].shape,grb)
        ### 根据hdf5_name创建一个H5文件对象
        result = hdf5_write_ndaray(h5path=hdf5_name,var_name=variable_name,level_name=tmp_var_level,ndarray_obj=tmp[0])
        print(result)



#####################################################################################################################################################################
#####################################################################################################################################################################





from meoteofence.interface import data_drill_up
from scipy.interpolate import interp1d
import numpy as np 
import pandas as pd



### 根据变量和层级对指定文件夹下的数据进行上钻操作
tmp_concat_df_a = data_drill_up(hdf5_store_path_pre='./data_handled/20221123/',
                                 variable_name = 'Wind speed (gust)',
                                 level_name = 'level_0',
                                 forecast_time = 24)
tmp_concat_df_b = data_drill_up(hdf5_store_path_pre='./data_handled/20221124/',
                                 variable_name = 'Wind speed (gust)',
                                 level_name = 'level_0',
                                 forecast_time = 24)
tmp_concat_df = pd.concat([tmp_concat_df_a,tmp_concat_df_b],axis=0)
tmp_concat_df = tmp_concat_df.reset_index(drop=True)
# print(tmp_concat_df)   

def data_interpolate(tmp_concat_df,index_delta,kind):
    ### 创建一个空dataframe用来装插值后的变量列表
    tmp_interpolate_df = pd.DataFrame()
    ### 开始循环----dataframe列变量
    first_column_name = tmp_concat_df.columns[0]
    ### 选取变量
    tmp_array = tmp_concat_df[first_column_name].values
    # print(tmp_array,type(tmp_array))
    ### 确定原本的x值
    tmp_original_x = np.array(list(tmp_concat_df.index))
    for tmp_column_name in tmp_concat_df.columns:
        # print(tmp_column_name)
        ### 对数据进行插值
        ### 拟合曲线
        tmp_func = interp1d(tmp_original_x,tmp_array,kind=kind)
        ### 用拟合好的曲线在新的序列上计算新的值
        start_index = tmp_concat_df.index[0]
        end_index = tmp_concat_df.index[-1]
        # index_delta = 0.25
        tmp_new_x = np.arange(start_index,end_index+index_delta,index_delta)
        tmp_new_array = tmp_func(tmp_new_x)
        tmp_interpolate_df[tmp_column_name] = list(tmp_new_array)
    return tmp_interpolate_df

tmp_interpolate_df = data_interpolate(tmp_concat_df=tmp_concat_df,index_delta=0.25,kind='linear')
tmp_target_df = tmp_interpolate_df.iloc[:96,:]
print(tmp_target_df)







from setuptools import setup,find_packages

### 打开readme文件流，使用utf-8编码
with open("README.mdown","r",encoding="utf-8") as fh:
        long_description = fh.read()

setup(
        ### 包与作者信息
        name = 'shihua-meoteofence',
        version = '0.1',
        author = 'shihua',
        author_email = "15021408795@163.com",
        python_requires = ">=3.9.12",
        license = "MIT",

        ### 源码与依赖
        packages = find_packages(),
        include_package_data = True,
        description = 'Meoteofence is a preprocessing program for numerical weather forecasting, which is mainly used to segment the original meteorological data according to the five dimensions of starting time, prediction time, spatial hierarchy, meteorological physical quantities and station coordinate grid. The main technology uses HDF5 to organize the matrix data after analysis, uses hook technology to organize the code, and adds parallel processing at the prediction time level.',
        # install_requires = ['pluggy','tables','pandas','numpy'],

        ### 包接入点，命令行索引
        # entry_points = {
        #     'console_scripts': [
        #         'fichectl = fiche.cli:fiche'
        #     ]
        # }      

        ### pypi配置
        long_description = long_description,
        long_description_content_type = "text/markdown",
        url = "https://github.com/redblue0216/Meoteofence",
        classsifiers = [
                "Programming Language :: Python :: 3.9",
                "License :: OSI Approved :: MIT License",
                "Topic :: Scientific/Engineering :: Mathematics"
        ]

)


# -*- coding: utf-8 -*-
# author:shihua
# coder:shihua
# 这是一个hook技术具体实现类，主要功能提供气象数据预处理具体实现，主要技术hook
"""
模块介绍
-------

这是一个hook技术具体实现类，主要功能提供气象数据预处理具体实现，主要技术hook

设计模式：

    （1）  无 

关键点：    

    （1）hook_impl

主要功能：            

    （1）hook技术具体实现                             

使用示例
-------


类说明
------

"""



####### 载入程序包 ##########################################################
############################################################################



import pluggy
import meoteofence
from meoteofence import get_initial_info_from_file,convert_matrix_to_list,get_coordinate_pairs_according_to_station,form_base_coordinate,combine_variable_and_levels,hdf5_write_ndaray,hdf5_read_ndarray,obtain_data_and_cache_to_hdf5
import pygrib 
import numpy as np
import os
import time



####### hook技术具体实现类 ###################################################
### 设计模式：                                                           ###
### （1）无                                                              ###
### 关键点：                                                             ###
### （1）hook_impl                                                      ###
### 主要功能：                                                           ###
### （1）hook技术具体实现                                                 ###
############################################################################


####### hook技术具体实现类 #####################################################################################
#############################################################################################################



hook_impl = pluggy.HookimplMarker(meoteofence.HOOK_NAMESPACE)



class PretreatmentPlugin(object):
    '''
    类介绍：

        这是一个预处理具体实现
    '''


    @hook_impl
    def meoteo_pretreatment(self,file_path,variable_name,station_location,hdf5_store_path_pre,resolution_ratio,grid_size):
        '''
        方法功能：

            定义一个气象预处理的具体实现方法

        参数：
            file_path (str): 文件路径
            variable_name (str): 变量名称
            station_location (tuple): 场站坐标元组
            hdf5_store_path_pre (str): hdf5存储路径前缀
            resolution_ratio (float): 分辨率
            grid_size (int): 网格大小

        返回：
            result (str): 运行结果信息
        '''

        ### 获取经纬度列表和坐标点位
        locations,variable_level_list,grbs_obj = get_initial_info_from_file(file_path=file_path,variable_name=variable_name)
        ### 将经纬度坐标转换为两个列表
        latitude_list,longitude_list,longtitude_latitude_list = convert_matrix_to_list(locations=locations)
        # print(latitude_list,longitude_list,longtitude_latitude_list)
        ### 以场站坐标为依据从两个列表中选取坐标对，示例使用longtitude-31.23,latitude-90.24
        # station_location = (31.23,90.24)
        upleft_standard_location,resolution_ratio = get_coordinate_pairs_according_to_station(longitude_list=longitude_list,latitude_list=latitude_list,station_location=station_location)
        # print(upleft_standard_location,resolution_ratio)
        ### 对应形成坐标矩阵
        upleft_grid_location,upright_grid_location,downleft_grid_location,downright_grid_location= form_base_coordinate(upleft_standard_location=upleft_standard_location,resolution_ratio=resolution_ratio,grid_size=grid_size)
        # print(upleft_grid_location,upright_grid_location,downleft_grid_location,downright_grid_location)
        ### 组合变量与层级得到宽表字段
        variable_level_combine_list = combine_variable_and_levels(variable_name=variable_name,variable_level_list=variable_level_list)
        # print(variable_level_combine_list)
        ### 获取数据并缓存到HDF5中
        ### 配置HDF5文件路径
        hdf5_file_path = file_path.split('.')[-2]+'.'+'h5'
        hdf5_file_path = hdf5_store_path_pre + hdf5_file_path
        # print(hdf5_file_path)
        ### 开始向HDF5缓存
        obtain_data_and_cache_to_hdf5(grbs_obj=grbs_obj,
                                    variable_name=variable_name,
                                    variable_level_list=variable_level_list,
                                    hdf5_name=hdf5_file_path,
                                    upleft_grid_location=upleft_grid_location,
                                    upright_grid_location=upright_grid_location,
                                    downleft_grid_location=downleft_grid_location,
                                    downright_grid_location=downright_grid_location)    
        print('======================================================>>>>>>',file_path,hdf5_file_path)
        result = "meoteo-pretreatment well done!"
        print(result)

        return result



###############################################################################################################################################
###############################################################################################################################################


